#!/usr/bin/env python
# coding: utf-8

from copy import copy, deepcopy

from nopea.exceptions import TooManyResultsError


class QuerySet:

    def __init__(self, base):
        self.base = base
        self.adaptor = self.base.adaptor
        self.query = ''
        self.query_args = []
        self.partials = {
            'filters': [],
            'excludes': [],
            'updates': [],
            'orders': [],
            'offset': None,
            'limit': None,
            'create': None,
            'all': False,
            'count': False,
            'delete': False,
        }

    def __repr__(self):
        result = self()
        appendix = ""
        if len(result) > 20:
            appendix = " ..remainung elements truncated."
        return "%s%s" % (str([item for item in result[:20]]), appendix)

    def __len__(self):
        result = self()
        return len(result)

    def make_objects(self, instance, db_results) -> list:
        objects = []
        cls = instance.__class__
        for result in db_results:
            _object = cls()
            mapped = zip(instance.fields, result)
            for field, value in mapped:
                field.make_value(value)
                setattr(_object, field.fieldname, field.value)

            for reverse_name, related_manager in cls.related_managers.items():
                new_manager = copy(related_manager)
                new_manager.id = _object.id
                setattr(_object, reverse_name, new_manager)

            objects.append(_object)

        return objects

    def _make_query_base(self):
        if self.partials['count'] == True:
            self.query = self.adaptor.get_count_query(self.base)
        else:
            self.query = self.adaptor.get_select_query(self.base)

    def _make_where_clause(self):
        if self.partials['filters'] or self.partials['excludes']:
            self.query += ' WHERE '

    def _make_updates(self):
        if self.partials['updates']:
            self.query, update_args = self.adaptor.get_update_query(self.base, self.partials['updates'][0])
            self.query_args.extend(update_args)
            if self.partials['filters']:
                self.query += 'WHERE '

    def _make_deletions(self):
        if self.partials['delete']:
            self.query = self.adaptor.get_delete_query(self.base)
            if self.partials['filters']:
                self.query += 'WHERE '

    def _make_filters(self):
        if self.partials['filters']:
            filters = []
            for filter_partial in self.partials['filters']:
                filter_query, filter_args = self.adaptor.get_filter_query(filter_partial)
                filters.append(filter_query)
                if isinstance(filter_args, list):
                    self.query_args.extend(filter_args)
                else:
                    self.query_args.append(filter_args)
            self.query += ' AND '.join([item for item in filters])
            if self.partials['excludes']:
                self.query += ' AND '

    def _make_excludes(self):
        if self.partials['excludes']:
            excludes = []
            for exclude_partial in self.partials['excludes']:
                exclude_query, exclude_args = self.adaptor.get_exclude_query(exclude_partial)
                excludes.append(exclude_query)
                if isinstance(exclude_args, list):
                    self.query_args.extend(exclude_args)
                else:
                    self.query_args.append(exclude_args)
            self.query += ' AND '.join([item for item in excludes])

    def _make_orders(self):
        if self.partials['orders']:
            self.query += ' ORDER BY '
            orders = []
            for order_partial in self.partials['orders']:
                direction = 'DESC' if order_partial[0].startswith('-') else 'ASC'
                orders.append('%s %s' % (order_partial[0].replace('-', ''), direction))
            self.query += ', '.join(orders)

    def _make_offset(self):
        if self.partials['offset']:
            self.query += self.adaptor.get_offset_query(self.partials['offset'], self.partials['limit'])

    def _make_limit(self):
        if self.partials['limit']:
            self.query += self.adaptor.get_limit_query(self.partials['limit'], self.partials['offset'])

    def compile_query(self) -> tuple:
        self.query_args = []

        self._make_query_base()
        self._make_where_clause()
        self._make_updates()
        self._make_deletions()
        self._make_filters()
        self._make_excludes()
        self._make_orders()
        self._make_limit()
        self._make_offset()
        return (self.query, self.query_args)

    def execute_sql(self):
        query, query_args = self.compile_query()
        result = self.adaptor.execute_query(query, query_args)[0]
        return result

    def __call__(self):
        result = self.execute_sql()
        if self.partials['count']:
            return result[0][0]
        return self.make_objects(self.base, result)

    def __iter__(self):
        objects = self.make_objects(self.base, self.execute_sql())
        for item in objects:
            yield item

    def __getitem__(self, s):
        if isinstance(s, int):
            start = s
            stop = s + 1

            self.partials['offset'] = start
            self.partials['limit'] = stop
            return list(self)[0]
        if isinstance(s, slice):
            start = s.start
            stop = s.stop

        self.partials['offset'] = start
        self.partials['limit'] = stop
        return list(self)

    def filter(self, *args, **kwargs):
        if not kwargs:
            self.partials['all'] = True
        for key, value in kwargs.items():
            self.partials['filters'].append({key: value})
        return self

    def exclude(self, *args, **kwargs):
        for key, value in kwargs.items():
            self.partials['excludes'].append({key: value})
        return self

    def all(self):
        return self.filter()

    def get(self, *args, **kwargs):
        result = self.filter(*args, **kwargs)()

        if len(result) <= 1:
            try:
                return result[0]
            except IndexError:
                raise IndexError("No %s object found" % self.base.__class__.__name__)
        raise TooManyResultsError('To many results for get(): %s' % len(result))

    def delete(self, *args, **kwargs) -> None:
        self.partials['delete'] = True
        self()
        return None

    def update(self, *args, **kwargs):
        count = 0
        filters = self.partials['filters']
        if filters:
            count_qs = QuerySet(self.base)
            for item in filters:
                count_qs = count_qs.filter(**item)
            count = len(count_qs)

        for key, value in kwargs.items():
            self.partials['updates'].append({key: value})
        self()
        return count

    def order_by(self, *args):
        self.partials['orders'].append(args)
        return self

    def first(self):
        self.partials['limit'] = 1
        result = self()
        if result:
            return result[0]

    def last(self):
        count_qs = QuerySet(self.base)
        filters = self.partials['filters']
        for item in filters:
            count_qs = count_qs.filter(**item)
        count = len(count_qs)

        self.partials['offset'] = count - 1
        result = self()
        if result:
            return result[0]

    def count(self):
        self.partials['count'] = True
        return self()

    def exists(self):
        return len(self) > 0



#!/usr/bin/env python
# coding: utf-8


class TooManyResultsError(Exception):
    error_code = 1
    pass


class MaxLengthError(Exception):
    error_code = 1
    pass


class UnknownFieldError(Exception):
    error_code = 1
    pass



#!/usr/bin/env python
# coding: utf-8

from nopea.exceptions import MaxLengthError
from nopea.managers import RelatedManager


class Unresolved:
    pass


class DbField(object):
    def __new__(cls, *args, **kwargs):
        if len(args) > 0:
            cls.adaptor = args[0].adaptor
            cls.fieldname = cls.__name__.lower()
        return super().__new__(cls)

    def __init__(self):
        self.value = Unresolved()

    def make_value(self, value):
        self.value = value

    def get_value(self):
        return self.value

    def get_drop_column_query(self, base=None):
        # Must be a function for sqlite
        return self.adaptor.get_drop_column_query(self, base)


class PkField(DbField):

    def __init__(self):
        self.fieldname = 'id'

    @property
    def partial_create_table_query(self):
        return self.adaptor.get_pkfield_create_query()


class IntegerField(DbField):

    def __init__(self, default=None):
        self.default = default

    @property
    def partial_create_table_query(self):
        return self.adaptor.get_integer_field_create_query(self)

    @property
    def add_column_query(self):
        return self.adaptor.get_integer_field_create_column_query(self)


class CharField(DbField):

    def __init__(self, max_length, default=None):
        if not isinstance(max_length, int) or max_length < 1:
            raise MaxLengthError("On CharFields max_length must be set and it must be a positive integer")
        self.max_length = max_length
        self.default = default

    @property
    def partial_create_table_query(self):
        return self.adaptor.get_char_field_create_query(self)

    @property
    def add_column_query(self):
        return self.adaptor.get_charfield_create_column_query(self)


class TextField(DbField):

    def __init__(self, default=""):
        self.default = default

    @property
    def partial_create_table_query(self):
        return self.adaptor.get_text_field_create_query(self)

    @property
    def add_column_query(self):
        return self.adaptor.get_text_field_create_column_query(self)


class BooleanField(DbField):

    def __init__(self, default):
        self.default = default

    @property
    def partial_create_table_query(self):
        return self.adaptor.get_boolean_field_create_query(self.default)

    @property
    def add_column_query(self):
        return self.adaptor.get_boolean_field_create_column_query(self)


class DateTimeField(DbField):

    def __init__(self, default=None):
        self.default = default

    @property
    def partial_create_table_query(self):
        return self.adaptor.get_datetime_field_create_query(self.default)

    @property
    def add_column_query(self):
        raise NotImplementedError("Still not implemented")


class ForeignKey(DbField):

    def __init__(self, reference_class, reverse_name=None, lazy=False, on_delete=''):
        self.reference_class = reference_class
        self.reverse_name = reverse_name
        self.lazy = lazy
        self.on_delete = on_delete

    def set_related_manager(self, target):
        if not self.reverse_name:
            self.reverse_name = target.__class__.__name__.lower() + '_set'
        setattr(self.reference_class, self.reverse_name, RelatedManager(target, self.fieldname, self.reverse_name))

    def make_value(self, value):
        if not self.lazy:
            if value:
                reference_obj = self.reference_class.objects.filter(id=value)
                if not reference_obj.exists():
                    self.value = None
                self.value = reference_obj[0]
        else:
            self.value = ReverseLazy(value, self.reference_class)

    def __get__(self, instance, owner):
        return self

    def get_value(self):
        if self.lazy:
            return self.value().id
        return self.value

    @property
    def partial_create_table_query(self):
        return self.adaptor.get_foreignkey_field_create_query_base()

    @property
    def partial_create_table_query_extension(self):
        return self.adaptor.get_foreignkey_field_create_query_extension(self.reference_class, self.on_delete)

    @property
    def add_column_query(self):
        raise NotImplementedError("Still not implemented")


class ByteField(DbField):

    @property
    def partial_create_table_query(self):
        return self.adaptor.get_byte_field_field_create_query(self)

    @property
    def add_column_query(self):
        return self.adaptor.get_byte_field_create_column_query(self)


class ReverseLazy:
    def __init__(self, value, reference_class):
        self.value = value
        self.reference_class = reference_class

    def __call__(self):
        if self.value:
            reference_obj = self.reference_class.objects.filter(id=self.value)
            if not reference_obj.exists():
                return None
            return reference_obj[0]
        return self



# This file is needed to initialize the models and migrations
import os
from nopea.dbobject import DbObject
from nopea.adaptors.sqlite import SQLiteAdaptor
from nopea.migrations import Migration

from models import *

DbObject.adaptor = SQLiteAdaptor('nopea.db')

# Examples for other DBMS

# MySQL
# from nopea.adaptors.mysql import MySQLAdaptor
# DbObject.adaptor = MySQLAdaptor({
#     'host': 'localhost',
#     'user': 'nopea',
#     'db': 'nopea',
#     'use_unicode': True,
#     'charset': 'utf8'
# })


# from nopea.adaptors.postgres import PostgreSQLAdaptor
# DbObject.adaptor = PostgreSQLAdaptor({
#     'host': 'localhost',
#     'user': 'nopea',
#     'database': 'nopea',
#     'password': 'nopea'
# })

Migration.migration_dir = os.path.join(os.getcwd(), 'migrations')



#!/usr/bin/env python
# coding: utf-8

from nopea.queryset import QuerySet
from nopea.exceptions import TooManyResultsError, UnknownFieldError


def check_fieldnames(func):
    def wrapper(*args, **kwargs):
        for field in kwargs.keys():
            field = field.split('__')[0]
            if field not in args[0].base.fieldnames:
                raise UnknownFieldError(f"Table {args[0].base.tablename} has no field {field}")
        return func(*args, **kwargs)
    return wrapper


class Manager:

    def __init__(self, base):
        self.base = base
        self.adaptor = self.base.adaptor

    @check_fieldnames
    def filter(self, *args, **kwargs):
        query_set = QuerySet(self.base)
        return query_set.filter(*args, **kwargs)

    @check_fieldnames
    def exclude(self, *args, **kwargs):
        query_set = QuerySet(self.base)
        return query_set.exclude(*args, **kwargs)

    def all(self):
        return self.filter()

    def first(self):
        return self.all().first()

    def last(self):
        return self.all().last()

    def count(self):
        return self.filter().count()

    def exists(self):
        return self.count() > 0

    @check_fieldnames
    def create(self, *args, **kwargs):
        query, query_args = self.adaptor.get_insert_query(self.base, kwargs)
        _, _id = self.adaptor.execute_query(query, query_args)
        return self.get(id=_id)

    def get(self, *args, **kwargs):
        result = self.filter(*args, **kwargs)()

        if len(result) <= 1:
            try:
                return result[0]
            except IndexError:
                raise IndexError("No %s object found" % self.base.__class__.__name__)
        raise TooManyResultsError('To many results for get(): %s' % len(result))

    @check_fieldnames
    def get_or_create(self, *args, **kwargs):
        try:
            return (False, self.get(*args, **kwargs))
        except IndexError:
            return (True, self.create(*args, **kwargs))

    def bulk_create(self, objects):
        self.adaptor.bulk_create(objects, self.base)


class RelatedManager(Manager):

    def __init__(self, target, fieldname, reverse_name):
        self.id = None
        self.base = target
        self.fieldname = fieldname
        target.related_managers[reverse_name] = self

    def filter(self, *args, **kwargs):
        field_filter = {self.fieldname: self.id}
        return super().filter(**field_filter).filter(*args, **kwargs)

    def exclude(self, *args, **kwargs):
        field_filter = {self.fieldname: self.id}
        return super().filter(**field_filter).exclude(*args, **kwargs)

    def bulk_create(self, *args, **kwargs):
        raise NotImplementedError("bulk_create cannot be used on RelatedManagers")



# coding: utf-8
from setuptools import setup

with open('README.md') as f:
    long_description = f.read()

setupargs = {
    'name': 'nopea',
    'description': 'Provides an ORM for MySQL, PostgreSQL and SQLite.',

    'license': 'GPLv3',
    'version': '0.0.21',

    'packages': ['nopea', 'nopea.adaptors'],
    'long_description': long_description,
    'long_description_content_type': 'text/markdown',

    'author': 'Christian Kokoska',
    'author_email': 'info@softcreate.de',
    'install_requires': [
        'colorama==0.3.9',
        'ipython'
    ],
}

if __name__ == '__main__':
    setup(**setupargs)
